/* 
   - The contents of this file are subject to the Mozilla Public
   - License Version 1.1 (the "License"); you may not use this file
   - except in compliance with the License. You may obtain a copy of
   - the License at http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS
   - IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
   - implied. See the License for the specific language governing
   - rights and limitations under the License.
   -
   - The Original Code is this file as it was released on
   - August 21, 2003.
   -
   - The Initial Developer of the Original Code is HJ van Rantwijk.
   - Portions created by HJ van Rantwijk are Copyright (C) 2003-2004
   - HJ van Rantwijk.  All Rights Reserved.
   -
*/

const NO_ROWS_AVAILABLE      = -1;
const NO_ROWS_SELECTED       = 0;
const SINGLE_ROW_SELECTED    = 1;
const MULTIPLE_ROWS_SELECTED = 2;
const ROW_IS_CONTAINER       = 4;
const MIXED_ROWS_SELECTED    = 8;

const nsIWindowDataSource = Components.interfaces.nsIWindowDataSource;

var pref = null; // declared in navigator.js but out of reach for us!
var gResource = "";

var prefWindow = null;
var prefPanelNames;
var managerTypes = [ "tabManager" ];
var dialogType = "tabManagerPrefWindow";
var panelPath = "chrome://multiviews/content/tab-manager/preferences/";
var gUseStackedWindows = false;
var resourcePrefix = "http://multizilla.mozdev.org/rdf#";

/***
  * Toolbar button ID arrays that will be replaced with elements in updateToolbarButtons()
  * note: keep the order in-line with the corresponding XUL code!
  */
var toolbarButtons = [ "tabButton", "windowButton", "sessionButton", "reloadButton", "propertiesButton", "closeButton", "prefsButton" ];
/***
  * Context Menu ID array that will be replaced with elements in initContextMenu(aEvent)
  * note: we have to keep the order in-line with the corresponding XUL code!
  */
var contextMenuItems = [ "doCmd_SelectAll", "doCmd_UndoSelection", "SetInterval" ];
/***
  * Toolbar button arrays, to ease the pain of enabling/disabling the buttons
  */
var defaultEnabledButtons      = [  true,  true, false, false, false, false, true ];
var selectionEnabledButtons    = [  true,  true,  true,  true,  true,  true, true ];
var multipleRowsEnabledButtons = [ false,  true, false, false, false,  true, true ];

const TAB_BUTTON        = 0;
const WINDOW_BUTTON     = 1;
const SESSION_BUTTON    = 2;
const INTERVAL_BUTTON   = 3;
const PROPERTIES_BUTTON = 4;
const CLOSE_BUTTON      = 5;
const QPREFS_BUTTON     = 6;

const CM_SELECT_ALL  = 0;
const CM_SELECT_NONE = 1;
const CM_DEFAULT     = 2;
const CM_CUSTOM      = 3;
const CM_ADD         = 4;
const CM_REPLACE     = 5;
const CM_EDIT        = 6;
const CM_REMOVE      = 7;

// Our global stuff goes here
var gTabbrowser = null;
var stringBundle = null;
var busyTabs = -1;
// Global vars used by confirmation dialog
var gConfirmCheckbox = false;
var gEscapeKeyPressed = false;
var gConfirmForID = 0;

var tree       = null;
var dataSource = null;

////////////////////////////////////////////////////////////////////////////////
function getNavigatorWindow(aOpenFlag)
{
  var windowManager = Components.classes['@mozilla.org/appshell/window-mediator;1'].getService();
  var windowManagerInterface = windowManager.QueryInterface(Components.interfaces.nsIWindowMediator);
  var navigatorWindow = windowManagerInterface.getMostRecentWindow("navigator:browser");

  if (!navigatorWindow) {
    if (!aOpenFlag)
      return null;
    else {
      window.open();
      navigatorWindow = windowManagerInterface.getMostRecentWindow("navigator:browser");
    }
  }
  return navigatorWindow;
}
////////////////////////////////////////////////////////////////////////////////
function getTabbrowser(aOpenFlag)
{
  var navigatorWindow = getNavigatorWindow(aOpenFlag);

  if (navigatorWindow)
    return navigatorWindow.getBrowser();
  return null;
}
////////////////////////////////////////////////////////////////////////////////
function areAllTabsBlank(aTabbrowser)
{
  var numberOfTabs = aTabbrowser.mTabContainer.childNodes.length;
  var numberOfBlankTabs = getNumberOfBlankTabs(aTabbrowser);

  return (numberOfTabs == numberOfBlankTabs);
}
////////////////////////////////////////////////////////////////////////////////
function getNumberOfBlankTabs(aTabBrowser)
{
  var blankTabs = 0;
  var browsers = aTabBrowser.mPanelContainer.childNodes.length;

  for (var i = 0; i < browsers; i++) {
    var browser = aTabBrowser.getBrowserAtIndex(i);
    var tab = aTabBrowser.getTabForBrowser(browser);

    if (tab.hasAttribute("busy")) {
      if (tab.getAttribute("firstPage") == "about:blank")
        blankTabs++;
    }
    else if (browser.currentURI.spec == "about:blank")
      blankTabs++;
  }
  return blankTabs;
}
////////////////////////////////////////////////////////////////////////////////
function getBrowser()
{
  return getTabbrowser(true);
}
////////////////////////////////////////////////////////////////////////////////
// This function is copied over from multiviewsInit.js because it is not 
// available from multiviewsRDF.js (read 'it is not included').
////////////////////////////////////////////////////////////////////////////////
function getNumberOfWindowsByType(aType)
{
  var defaultType = "navigator:browser";

  if (!aType) // default check for browser window
    aType = defaultType;

  var windowCounter = 0;
  var windowManager = Components.classes['@mozilla.org/appshell/window-mediator;1'].getService(Components.interfaces.nsIWindowMediator);
  var enumerator = windowManager.getEnumerator(aType);

  while (enumerator.hasMoreElements()) {
    var aWindow = enumerator.getNext();
    windowCounter++;
  }
  return windowCounter;
}
////////////////////////////////////////////////////////////////////////////////
function updateToolbarButtons()
{
  // Is the toolbar button array initialized?
  if (typeof toolbarButtons[0] == 'string') {
    // Nope, not yet, initialize array now
    for (i in toolbarButtons) {
      toolbarButtons[i] = document.getElementById(toolbarButtons[i]);
    }
  }
  var buttons = defaultEnabledButtons;
  var treeState = tabManager.getCurrentTreeState();

  if (treeState > NO_ROWS_SELECTED) {
    if (treeState & SINGLE_ROW_SELECTED) {
      buttons = selectionEnabledButtons;
    }
    else if (treeState == MULTIPLE_ROWS_SELECTED)
      buttons = multipleRowsEnabledButtons;
  }
  // Enable/disable the toolbar buttons based on the toolbar array
  for (i in toolbarButtons) {
    toolbarButtons[i].disabled = !buttons[i];
  }
  // Additional checks to disable one or more toolbar buttons
  var isLastBrowserWindow = (getNumberOfWindowsByType("navigator:browser") == 1);
  var disabled = ((treeState == NO_ROWS_SELECTED && !isLastBrowserWindow)|| 
                  treeState & MULTIPLE_ROWS_SELECTED); 
  toolbarButtons[TAB_BUTTON].disabled = disabled;
  toolbarButtons[SESSION_BUTTON].disabled = (treeState == NO_ROWS_SELECTED || 
                                            (treeState == SINGLE_ROW_SELECTED && !(treeState & ROW_IS_CONTAINER)) || 
                                            (treeState == MULTIPLE_ROWS_SELECTED && !(treeState & ROW_IS_CONTAINER)));
  disabled = (treeState == NO_ROWS_SELECTED || 
              treeState == MULTIPLE_ROWS_SELECTED || 
              treeState & ROW_IS_CONTAINER);
  toolbarButtons[INTERVAL_BUTTON].disabled = disabled;
  toolbarButtons[PROPERTIES_BUTTON].disabled = disabled;
  toolbarButtons[CLOSE_BUTTON].disabled = (treeState == NO_ROWS_SELECTED || (treeState & ROW_IS_CONTAINER && isLastBrowserWindow));
}
////////////////////////////////////////////////////////////////////////////////
function initContextMenu(aEvent)
{
  // Is the context menu item array initialized?
  if (typeof contextMenuItems[0] == 'string') {
    // Nope, not yet, initialize array now
    for (i in contextMenuItems) {
      contextMenuItems[i] = document.getElementById(contextMenuItems[i]);
    }
  }
  for (i in contextMenuItems) {
    contextMenuItems[i].setAttribute('disabled', 'true');
  }
  var treeState = tabManager.getCurrentTreeState();

  if (treeState == NO_ROWS_AVAILABLE)
    return;

  var treeRows = tree.treeBoxObject.view.rowCount;

  if (treeState == MULTIPLE_ROWS_SELECTED) {
    contextMenuItems[CM_SELECT_ALL].setAttribute('disabled', 'true');
    contextMenuItems[CM_SELECT_NONE].removeAttribute('disabled');
  }
  else { // NO_ROWS_SELECTED || SINGLE_ROW_SELECTED
    if (treeRows > 1)
      contextMenuItems[CM_SELECT_ALL].removeAttribute('disabled');
    contextMenuItems[CM_SELECT_NONE].setAttribute('disabled', 'true');
  }
}
//////////////////////////////////////////////////////////////////////////////
function initReloadPopup(aPopup, aEvent)
{
  if (toolbarButtons[3].disabled) {
    aEvent.preventDefault();
    return;
  }
  gTabbrowser.initReloadPopup(aPopup, null, tabManager.getReloadInterval());
}
////////////////////////////////////////////////////////////////////////////////
function openTMPropertyWindow()
{
  var treeState = tabManager.getCurrentTreeState();

  if (treeState == NO_ROWS_SELECTED || treeState == MULTIPLE_ROWS_SELECTED || currentRow == -1)
    return;
  gResource = tabManager.getResourceForRow(currentRow); // tab    -> urn:session-x:window-x:tabs:tab-x
                                                            // window -> urn:session-x:window-x:tabs:root
  var winFeatures = "chrome, dependent, modal, dialog" + (gUseStackedWindows ? "" : ", resizable=yes, centerscreen");
  window.openDialog("chrome://multiviews/content/tab-manager/properties/tabManagerProperties.xul", 
                    "tmproperties", winFeatures, window);
}
////////////////////////////////////////////////////////////////////////////////
function openPrefOverlay()
{
  var winFeatures = "chrome, dependent, modal, dialog" + (gUseStackedWindows ? "" : ", resizable=yes, centerscreen");
  window.openDialog("chrome://multiviews/content/tab-manager/preferences/tabManagerPrefWindow.xul", 
                    "prefWindow", winFeatures, window, "");
}
////////////////////////////////////////////////////////////////////////////////
// This is a dummy function, called from the pref manager (for compatibility only)
function switchTree(gPanel)
{
  return;
}
////////////////////////////////////////////////////////////////////////////////
function switchPanel(aPanel)
{
  var categoryButton = null;
  var panelFile = null;

  prefWindow.gSecondLevelPrefs = false;

  if (aPanel == "tabManagerPrefWindow")
    panelFile = "tabManagerMainPanel.xul";
  prefWindow.refreshPanel(panelFile, categoryButton);
}
////////////////////////////////////////////////////////////////////////////////
function initWindowOpenType()
{
  try {
    gUseStackedWindows = pref.getBoolPref("multizilla.managers.use-stacked-windows");
  } catch(ex) {
    gUseStackedWindows = /Win/.test(navigator.platform);
  }
}
////////////////////////////////////////////////////////////////////////////////
function getConfirmation(aWindow, aTitle, aText, aCheckboxText)
{
  gConfirmCheckbox = false;
  var checkbox = {};
  var checkboxText = aCheckboxText ? stringBundle.GetStringFromName(aCheckboxText) : null;
  var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
  var result = promptService.confirmEx(aWindow, stringBundle.GetStringFromName(aTitle), stringBundle.GetStringFromName(aText),
                                       (promptService.BUTTON_POS_0 * promptService.BUTTON_TITLE_YES) + 
                                       (promptService.BUTTON_POS_1 * promptService.BUTTON_TITLE_NO),
                                       null, null, null, checkboxText, checkbox);

  gConfirmCheckbox = checkbox.value;
  return (result == 0); // Yes returns 0 / No, Esc and X returns 1
}
////////////////////////////////////////////////////////////////////////////////
var tabManager = 
{
  windowManagerDS : null,
  tree : null,
  currentRow : null,
  stringBundle : null,
  reloadInterval : 0,

  initTabManager: function()
  {
    windowManagerDS = Components.classes['@mozilla.org/rdf/datasource;1?name=window-mediator'].getService(nsIWindowDataSource);  
    dataSource = Components.classes['@mozilla.org/rdf/datasource;1?name=in-memory-datasource'].createInstance(Components.interfaces.nsIRDFDataSource);
    tree = document.getElementById('tabManagerTree');
    tree.database.AddDataSource(dataSource);

    gTabbrowser = getTabbrowser(true);
    gBrowser = gTabbrowser; // this global is used in multiviewsRDF.js
    stringBundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
                             .getService(Components.interfaces.nsIStringBundleService)
                             .createBundle("chrome://multiviews/locale/multizilla.properties");
    currentRow = -1;
    reloadInterval = 0;
    var prefService = Components.classes["@mozilla.org/preferences-service;1"]
                                .getService(Components.interfaces.nsIPrefService);
    pref = prefService.getBranch(null);
    prefPanelNames = stringBundle.GetStringFromName("prefPanelNames").split(',');

    tabManager.buildTree();
    sizeToContent();
    initWindowOpenType();
   
    updateToolbarButtons();
  },

  buildTree: function()
  {
    var windowResources = getWindowResourceIDList();

    if (windowResources) {
      var sessionID = 0;

      for (i in windowResources) {
        var windowID = windowResources[i];
        var browserWindow = windowManagerDS.getWindowForResource(windowID);
        var browser = browserWindow.getBrowser();
        var resource = "urn:" + windowID + ":tabs:root";
        storeGroupData(dataSource, resource, browser, true, false);
      }
    }
  },

  getRowNumber: function(aEvent)
  {
    var X = aEvent.pageX;
    var Y = aEvent.pageY;
    return tree.treeBoxObject.getRowAt(X, Y);
  },

  getResourceForRow: function(aRowNumber)
  {
    mzDebug("getResourceForRow(" + aRowNumber + ")");
    return (aRowNumber > -1) ? tree.builderView.getResourceAtIndex(aRowNumber).Value : null;
  },

  getSelectedRows: function()
  {
    mzDebug("getSelectedRows()");
    var rows = -1;
    var selectedRows = new Array();
    // Get and return all selection ranges
    var rangeCount = tree.builderView.selection.getRangeCount();
    // Iterate over all ranges to collect the selected rows
    for (var range = 0; range < rangeCount; range++) {
      var startSelection = {};
      var endSelection = {};
      tree.builderView.selection.getRangeAt(range, startSelection, endSelection);
      // Collect all rows and store them in our array
      for (var row = startSelection.value; row <= endSelection.value; row++) {
        // Check for invalid row numbers
        if (startSelection.value != -1 || endSelection.value != -1)
          selectedRows[++rows] = row;
      }
    }
    if (selectedRows)
      return selectedRows;
    return null;
  },

  getCurrentTreeState: function()
  {
    mzDebug("getCurrentTreeState()");
    if (tree.treeBoxObject.view.rowCount < 1)
      return NO_ROWS_AVAILABLE;

    var selectedRows = this.getSelectedRows();

    if (selectedRows.length > 1) {
      if (this.isMixedSelection())
        return (MULTIPLE_ROWS_SELECTED + MIXED_ROWS_SELECTED);
	  else if (tree.treeBoxObject.view.isContainer(Number(selectedRows[0])))
        return (MULTIPLE_ROWS_SELECTED + ROW_IS_CONTAINER);
      return MULTIPLE_ROWS_SELECTED;
    }
    else if (selectedRows.length == 1) {
      if (tree.treeBoxObject.view.isContainer(Number(selectedRows[0])))
        return (SINGLE_ROW_SELECTED + ROW_IS_CONTAINER);
      return SINGLE_ROW_SELECTED;
    }
    return NO_ROWS_SELECTED;
  },

  getRowSource: function(row, inout)
  {
    mzDebug("getRowSource(" + row + ")");
    var resource = this.getResourceForRow(row);
    var windowID = resource.replace(/urn:/, '');

	if (tree.treeBoxObject.view.isContainer(row))
      windowID = windowID.replace(/:tabs:\w*/, '');
    else 
      windowID = windowID.replace(/:tabs:tab-\w*/, '');

    var browserWindow = windowManagerDS.getWindowForResource(windowID);
    dump("\nbrowserWindow: " + browserWindow);
    var tabbrowser = browserWindow.gBrowser;

    var columnID = "tm-ordinal";
    columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
    var ordinal = tree.view.getCellText(row, columnID);
    // dump("\nordinal: " + ordinal);
    var tab = tabbrowser.mTabList[ordinal];
    var browser = tabbrowser.getBrowserForTab(tab);

    inout.window = browserWindow;
    inout.windowID = windowID;
    inout.resource = resource;
    inout.tabbrowser = tabbrowser;
    inout.tab = tab;
    inout.browser = browser;
  },

  getReloadInterval: function(aEvent)
  {
    var interval = 0;
    var columnID = "tm-interval";
    columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;

    if (aEvent) {
      var row = this.getRowNumber(aEvent);
      interval = (row > -1) ? tree.view.getCellText(row, columnID) : 0;
    }
    else if (!aEvent && tree.currentIndex != -1)
      interval = tree.view.getCellText(tree.currentIndex, columnID);

    if (interval)
      return interval;
    return 1;
  },

  setReloadInterval: function(aEvent, aValue)
  {
    var resource = this.getResourceForRow(currentRow);
    var intervalProp = RDF.GetResource("http://multizilla.mozdev.org/rdf#interval");
    var currentValue = dataSource.GetTarget(RDF.GetResource(resource), intervalProp, true);

    if (aValue == 0) // Remove interval property from tree
      dataSource.Unassert(RDF.GetResource(resource), intervalProp, currentValue);
    else
      addOrChangeValue(dataSource, RDF.GetResource(resource), intervalProp, currentValue, RDF.GetLiteral(aValue));

    this.updateTab("interval", aValue);
  },

  isMixedSelection: function()
  {
    var tabs = 0;
    var windows = 0;
    var mixed = false;

    // Collect selected tree rows
    var selectedItems = this.getSelectedRows();
    // Do we have any selected items?
    if (selectedItems && selectedItems.length > 0) {
      // Iterate over the seleted items
      for (var row = 0; row < selectedItems.length; row++) {
        if (tree.treeBoxObject.view.isContainer(selectedItems[row]))
          windows++;
        else
          tabs++;
        // Do we have a mixed selection?
        if (tabs > 0 && windows > 0) {
          mixed = true;
          break;
        }
      }
    }
    return mixed;
  },

  removeWindowItemsFromSelection: function()
  {
    // Collect selected tree rows
    var selectedRows = this.getSelectedRows();
    // Do we have any selected items?
    if (selectedRows) {
      // Iterate over the selected items to clear the 'window' (container) items
      for (row in selectedRows) {
        if (tree.treeBoxObject.view.isContainer(row))
          tree.view.selection.clearRange(row, row);
      }
    }
  },

  selectRows: function(aCommand)
  {
    switch (aCommand) {
      case "all": // Called from the context menu, or Ctrl+A ?
           tree.view.selection.selectAll();
           this.removeWindowItemsFromSelection();
           break;
      case "none": // Called from the context menu, or Ctrl+Z ?
           tree.view.selection.clearSelection();
           // dump("\ntree.currentIndex: " + tree.currentIndex);
           tree.view.selection.select(tree.currentIndex);
           tree.treeBoxObject.ensureRowIsVisible(tree.currentIndex);
           break;
    }
    tree.focus();
  },

  activateSelectedTab: function(row)
  {
    mzDebug("activateSelectedTab(" + row + ")");
    var obj = new Object();
    this.getRowSource(row, obj);

    if (obj.resource.indexOf("root") != -1) {
      document.commandDispatcher.advanceFocusIntoSubtree(obj.tabbrowser.selectedTab);
    }
    else {
      obj.tabbrowser.selectedTab = obj.tab;
      // document.commandDispatcher.advanceFocusIntoSubtree(obj.tab);
      /* if (obj.tabbrowser.selectedTab == obj.tab)
        document.commandDispatcher.advanceFocusIntoSubtree(obj.tab);
        // document.commandDispatcher.focusedElement = obj.tab;
      else
        obj.tabbrowser.selectedTab = obj.tab; */
    }
    this.refreshAfterCommand();
  },

  add: function(openType)
  {
    mzDebug("add(" + openType + ")");

    if (openType == 'window') {
      window.open();
    }
    else {
      var treeState = this.getCurrentTreeState();

      if (treeState == NO_ROWS_SELECTED) {
        if (openType == 'tab')
          gTabbrowser.addTab();
      }
      else {
        var obj;

        if (openType == 'tab') {
          obj = new Object();
          this.getRowSource(currentRow, obj);
          obj.tabbrowser.addTab();	
          /* if (obj.resource.indexOf("root"))
            obj.tabbrowser.addTab();
          else 
            gTabbrowser.addTab(); */
        }
        else if (openType == 'session') {
          if (mzRDFStore == null)
            initializeRDFStorageFile();

          if (!this.isMixedSelection()) {
            // var selectedRows = this.getSelectedRows();
		  
            if (selectedRows) {
              var skippedWindows = 0;
              var totalWindows = 0;
              var priorTabbrowser = null;

              for (var i = 0; i < selectedRows.length; i++) {
                totalWindows++;
                obj = new Object();
                this.getRowSource(selectedRows[i], obj);
                // are all tabs 'about:blank'?
                if (areAllTabsBlank(obj.tabbrowser)) {
                  skippedWindows++;
                  continue; // Yes, skip this window
                }
                var sessionID = 0;
                var windowID = obj.windowID;
                sessionID = getNextSessionID(sessionID, windowID);
                var resource = "urn:" + sessionID + ":" + windowID + ":tabs:root";
                storeGroupData(mzRDFStore, resource, obj.tabbrowser);
              }
              if (skippedWindows) {
                var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
                var promptTitle = stringBundle.GetStringFromName("tmSkippedWindowPromptTitle");
                var argArray = [ skippedWindows, totalWindows ];
                var promptText = stringBundle.formatStringFromName("tmSkippedWindowPromptText", argArray, 2);
                promptService.alert(window, promptTitle, promptText);
              }
            }
          }
        }
        obj = null;
      }
    }
    this.refreshAfterCommand();
  },

  closeTabs: function(allTabs)
  {
    if (allTabs) // needed for context menu items, might be moved out
      tabManager.selectRows('all');
    closeTabsOrWindows();
  },

  closeWindows: function ()
  {
    closeTabsOrWindows();
  },

  closeSelectedItems: function()
  {
    var treeState = this.getCurrentTreeState();

    if (treeState == NO_ROWS_SELECTED)
      return;
    else {
      var confirmRequired = true;
      var confirmPrefID = "multizilla.tab-manager.confirm.close-items"; 
      // read the pref, and add it, with the default value, when it is missing
      try {
        confirmRequired = pref.getBoolPref(confirmPrefID);
      } catch(ex) {
        pref.setBoolPref(confirmPrefID, true);
      }
      if (confirmRequired) {
        var header = "tabs-windows";

        if (treeState & SINGLE_ROW_SELECTED) {
          header = (treeState & ROW_IS_CONTAINER) ? "window" : "tab";
        }
        else { // MULTIPLE_ROWS_SELECTED
          if (!(treeState & MIXED_ROWS_SELECTED))
            header = (treeState & ROW_IS_CONTAINER) ? "windows" : "tabs";
        }
        var confirmationTitle = header + ".confirmDialogTitle";
        var confirmationText = header + ".confirmDialogText";
        // Ask for the users confirmation before removing the tab(s)
        if (!getConfirmation(window, confirmationTitle, confirmationText, null)) {
          tree.focus();
          return;
        }
      }
      this.close();
    }
  },

  close: function()
  {
    mzDebug("close()");
    var selectedRows = this.getSelectedRows();

    if (selectedRows) {
      // Close the tab(s) or Window(s) one after one, from buttom up!
      for (var i = (selectedRows.length-1); i > -1; i--) {
        var obj = new Object();
        this.getRowSource(selectedRows[i], obj);
        // is this a window 'urn:window-x:tabs:root'
        if (obj.resource.indexOf("root") != -1) {
          obj.window.gConfirmForID = 4;
          obj.window.tryToClose();
        }
        else { // this must be a tab 'urn:window-x:tabs:tab-x'
          // is this the last tab?
          if (obj.tabbrowser.mTabContainer.childNodes.length > 1)
            removeResources(dataSource, obj.resource); // no, remove tab from tree

          obj.tabbrowser.removeTab(obj.tab);
        }
      }
      obj = null;
      mzFlushDataSource(dataSource);
      tree.builder.rebuild();
      this.refreshAfterCommand();
    }
    mzDebug("exit closeTabs()");
  },

  refreshAfterCommand: function()
  {
    /* var focusedWindow = document.commandDispatcher.focusedWindow;
    if (focusedWindow != window)
      dump("\nWrong Window Focused"); */
      
    updateToolbarButtons();
    // window.focus(); // focus TM Window, to make the ESC key work
    tree.focus(); // focus TM Tree, to make key up/down work
    document.commandDispatcher.focusedWindow = window;
    document.commandDispatcher.focusedElement = tree;
  },

  updateTab: function(updateType, data)
  {
    var obj = new Object();
    var treeState = this.getCurrentTreeState();
    this.getRowSource(treeState, obj);
    var tab = obj.tab;
    var browser = obj.browser;
    var tabbrowser = obj.tabbrowser;
    var columnID;

    mzDebug("updateType: " + updateType + "\ndata: " + data);

    switch(updateType) {
      case "label":
        tab.setAttribute("label", data);
        columnID = "Name";
        columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
        tree.builderView.setCellText(currentRow, columnID, data);
        break;
      case "customlabel":
        if (data)
          tab.setAttribute("customLabel", "true");
        else 
          tabbrowser.removeCustomLabel();
        // tree.builderView.setCellValue(currentRow, '', data);
        // tree.builderView.setCellText(currentRow, '', data);
        break;
      case "location":
        browser.loadURI(data);
        columnID = "tm-location";
        columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
        tree.builderView.setCellText(currentRow, columnID, data);
        break;
      case "interval":
        tabbrowser.setReloadInterval(tab, data);
        break;
      case "allowPlugins":
      case "allowJavascript":
      case "allowMetaRedirects":
      case "allowSubframes":
      case "allowImages":
        browser.docShell[updateType] = data;
        tabbrowser.refreshDocShellPropertyButtons();
        break;
    }
  },
  
  onSelect: function()
  {
    currentRow = tree.currentIndex;
    updateToolbarButtons();
    var treeState = this.getCurrentTreeState();

    if (currentRow != -1 && (treeState & SINGLE_ROW_SELECTED))
      this.activateSelectedTab(currentRow);
  },

  onHelp: function(aPageAnchor)
  {
    var pageAnchor = "";
    var helpPages = stringBundle.GetStringFromName("helpPages-TabManager");

    if (aPageAnchor) {
      if (typeof aPageAnchor == 'string')
        pageAnchor = "#" + aPageAnchor;
      else // event
        pageAnchor = "#" + aPageAnchor.target.id.replace(/^tabManager-/, '');
    }
    // We could eventually check the on-line state!
    var tabbrowser = getTabbrowser(true);
    tabbrowser.selectedTab = tabbrowser.addOrGetTab(helpPages + pageAnchor);
    window.close();
  },

  onCancel: function()
  {
    var retValue = this.confirmClose(gConfirmForID);
    return retValue;
  },

  confirmClose: function(aConfirm)
  {
    var shouldClose = true;
    var confirmPrefID = "multizilla.tab-manager.confirm.quit";
    var dialogCheckboxText, confirmPrefValue = 15;
    /* ::::
       1  = Close Button
       2  = ESC key pressed
       4  = Ctrl-W
       8  = Click on [X] on Window
    :::: */
    try {
      // try to read the pref
        confirmPrefValue = pref.getIntPref(confirmPrefID);
    } catch(ex) {
      // Yikes, no pref found, add it with the default value
      pref.setIntPref(confirmPrefID, confirmPrefValue);
    }
    // Is the pref set for confirmation?
    if (confirmPrefValue && (confirmPrefValue & aConfirm) == aConfirm) {
      // Yes, ask the users confirmation
      switch(aConfirm) {
        case (aConfirm & 1): dialogCheckboxText = 'managerSkipConfirmForCloseButton';
          break;
        case (aConfirm & 2): dialogCheckboxText = 'managerSkipConfirmForEscapeKey';
          break;
        case (aConfirm & 4): dialogCheckboxText = 'managerSkipConfirmForCtrlW';
          break;
        case (aConfirm & 8): dialogCheckboxText = 'managerSkipConfirmForWindowX';
          break;
      }
      shouldClose = getConfirmation(window, 'tmConfirmCloseTitle', 'tmConfirmCloseText', dialogCheckboxText);
      // Is the checkbox checked and do we have a pref value to work with?
      if (gConfirmCheckbox && (confirmPrefValue >= aConfirm)) {
        // Yes, update pref
        pref.setIntPref(confirmPrefID, (confirmPrefValue -= aConfirm));
      }
    }
    return shouldClose;
  }
};

addEventListener("load", tabManager.initTabManager, false);

/* 
////////////////////////////////////////////////////////////////////////////////
function updateCurrentIndex()
{
  tree.currentIndex = gTabbrowser.selectedTab.ordinal;
  tree.view.selection.toggleSelect(tree.currentIndex);
}

var tabManagerDNDObserver = 
{
  tree : null,
  
  initTabManagerDNDObserver: function()
  {
    tree = document.getElementById("tabManagerTree");
    initializeRDFStorageFile();
    tree.database.AddDataSource(mzRDFStore);
    tree.builder.rebuild();
    updateCurrentIndex();
    tabbrowser = getLastActiveBrowserWindow().getBrowser();
    stringBundle = document.getElementById("tm-properties");
    tabListInit();
  },

  onDragStart: function(aEvent, aXferData, aDragSession)
  {
    // start the row drag operation only when there are at least 2 visible rows
    if (tree.treeBoxObject.view.rowCount > 1) {
      aEvent.target.setAttribute("dragover", "true");
      var columnID = "sm-interval";
      columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
      var ordinal = tree.view.getCellText(tree.currentIndex, columnID);
      // dump("onDragStart ordinal: "+ordinal+"\n");
      aXferData.data = new TransferData();
      aXferData.data.addDataForFlavour("text/unicode", ordinal);
    }
  },

  onDragOver: function(aEvent, aXferData, aDragSession)
  {
    var X = aEvent.pageX;
    var Y = aEvent.pageY;
    var row = tree.treeBoxObject.getRowAt(X, Y);

    aDragSession.canDrop = true;
    if (row > -1) {
      dump("onDragOver - row: "+row+"\n");
      var columnID = "sm-interval";
      columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
      var ordinal = tree.view.getCellText(row, columnID);
      dump("ordinal: "+ordinal+"\n");
      ordinal++;
      var element = document.getElementById("tm-tab-"+ordinal);
      dump(element.nodeName+"\n");
      // nsDragAndDrop.mDragSession.sourceNode = element;
    }
  },

  onDrop: function(aEvent, aXferData, aDragSession)
  {
    dump("onDrop\n"+aEvent.originalTarget.nodeName);
  },

  onDragExit: function(aEvent, aXferData, aDragSession)
  {
    tree.childNodes[1].removeAttribute("dragover");
  },

  getSupportedFlavours: function ()
  {
    // we might want to expand this
    var flavourSet = new FlavourSet();
    flavourSet.appendFlavour("text/unicode");
    return flavourSet;
  },

  canDrop: function (aEvent, aDragSession)
  {
    aDragSession.canDrop = true;
    // allow drops from the Tab Manager only
    var doc = aDragSession.sourceNode.ownerDocument;
    var elem = doc.getElementById("tabManagerWindow");
    return (elem && (elem.getAttribute("windowtype") == "multizilla:tabmanager"));
  } 
};
addEventListener("load", tabManagerDNDObserver.initTabManagerDNDObserver, false); */
